from abc import ABCMeta, abstractmethod
from dataclasses import dataclass, field
from typing import TypeVar

from ludic.base import BaseElement

from .types import BaseSize, Color, ColorRange, Size, SizeClamp

highlight_light: str | type = "default"
highlight_dark: str | type = "default"
try:
    from .highlight import LudicDark, LudicLight

    highlight_light = LudicLight
    highlight_dark = LudicDark
except ImportError:
    pass

_T = TypeVar("_T", bound="BaseElement")


@dataclass
class Colors:
    """Colors for a theme."""

    primary: Color = ColorRange(["#276662", "#4ecdc4", "#dbf5f3"])
    secondary: Color = ColorRange(["#212529", "#414549", "#616569"])
    success: Color = ColorRange(["#637a32", "#c7f464", "#eefbd0"])
    info: Color = ColorRange(["#978801", "#fce303", "#fef9cc"])
    warning: Color = ColorRange(["#7e4801", "#fc9003", "#fee8cc"])
    danger: Color = ColorRange(["#711414", "#e32929", "#f9d4d4"])

    light: Color = Color("#f2f2f2")
    dark: Color = Color("#414549")

    white: Color = Color("#fff")
    black: Color = Color("#222")


@dataclass
class Header:
    """Header for a theme."""

    size: BaseSize = SizeClamp(1.5, 2, 2.5)
    anchor: bool = False


@dataclass
class Headers:
    """Headers for a theme."""

    h1: Header = field(
        default_factory=lambda: Header(size=SizeClamp(2, 1.5, 3.3), anchor=False)
    )
    h2: Header = field(
        default_factory=lambda: Header(size=SizeClamp(1.5, 1.3, 2.8), anchor=False)
    )
    h3: Header = field(
        default_factory=lambda: Header(size=SizeClamp(1.2, 0.8, 2.2), anchor=False)
    )
    h4: Header = field(
        default_factory=lambda: Header(size=SizeClamp(1, 0.6, 1.8), anchor=False)
    )
    h5: Header = field(
        default_factory=lambda: Header(size=SizeClamp(0.7, 0.5, 1.3), anchor=False)
    )
    h6: Header = field(
        default_factory=lambda: Header(size=SizeClamp(0.5, 0.4, 1.1), anchor=False)
    )


@dataclass
class Fonts:
    """Font sizes for a theme."""

    primary: str = "Helvetica Neue, Helvetica, Arial, sans-serif"
    secondary: str = "Georgia, serif"
    monospace: str = "Space Mono, Roboto Mono, monospace"

    size: BaseSize = Size(1, "em")


@dataclass
class Sizes:
    """Size for a theme."""

    xxxxs: BaseSize = SizeClamp(0.39, 0, 0.47)
    xxxs: BaseSize = SizeClamp(0.47, 0.1, 0.69)
    xxs: BaseSize = SizeClamp(0.57, 0.15, 0.69)
    xs: BaseSize = SizeClamp(0.69, 0.2, 0.84)
    s: BaseSize = SizeClamp(0.84, 0.25, 1)
    m: BaseSize = SizeClamp(1, 0.3, 1.16)
    l: BaseSize = SizeClamp(1.16, 0.35, 1.4)  # noqa
    xl: BaseSize = SizeClamp(1.4, 0.4, 1.68)
    xxl: BaseSize = SizeClamp(1.68, 0.45, 2.01)
    xxxl: BaseSize = SizeClamp(2.01, 0.5, 2.41)
    xxxxl: BaseSize = SizeClamp(2.41, 0.55, 2.88)


@dataclass
class Borders:
    """Border sizes for a theme."""

    thin: BaseSize = Size(0.1)
    normal: BaseSize = Size(0.23)
    thick: BaseSize = Size(0.42)


@dataclass
class Rounding:
    """Border rounding for a theme."""

    less: BaseSize = Size(0.20)
    normal: BaseSize = Size(0.25)
    more: BaseSize = Size(0.35)


@dataclass
class Sidebar:
    """Sidebar layout config for a theme."""

    # The width of the sidebar (empty means not set; defaults to the content width)
    side_width: str | None = None

    # The narrowest the content element can be before wrapping. Should be a percentage.
    content_min_width: str = "60%"


@dataclass
class Switcher:
    """Switcher layout config for a theme."""

    # The container width at which the component switches between a horizontal and
    # vertical layout
    threshold: BaseSize = Size(40, "rem")

    # The maximum number of elements allowed to appear in the horizontal configuration
    limit: int = 4


@dataclass
class Cover:
    """Cover layout config for a theme."""

    # The minimum height of the cover
    min_height: BaseSize = Size(100, "vh")

    # The minimum space between and around the child elements
    element: str = "h1"


@dataclass
class Grid:
    """Grid layout config for a theme."""

    # The size of the grid cells
    cell_size: BaseSize = Size(250, "px")


@dataclass
class Frame:
    """Frame layout config for a theme."""

    # The width of the frame as fraction
    numerator: int = 16

    # The height of the frame as fraction
    denominator: int = 9


@dataclass
class Layouts:
    """Layout configuration for a theme."""

    sidebar: Sidebar = field(default_factory=Sidebar)
    switcher: Switcher = field(default_factory=Switcher)
    cover: Cover = field(default_factory=Cover)
    grid: Grid = field(default_factory=Grid)
    frame: Frame = field(default_factory=Frame)


@dataclass
class CodeBlock:
    """Code block config for a theme."""

    color: Color = Color("#414549")
    background_color: Color = Color("#f2f2f2")
    line_number_color: Color = Color("#c2c2c2")
    line_numbers: bool = True
    font_size: BaseSize = Size(0.9)
    style: str | type = highlight_light


@dataclass
class Theme(metaclass=ABCMeta):
    """An abstract base class for theme configuration."""

    measure: BaseSize = Size(70, "ch")
    line_height: float = 1.5

    fonts: Fonts = field(default_factory=Fonts)
    colors: Colors = field(default_factory=Colors)
    sizes: Sizes = field(default_factory=Sizes)

    borders: Borders = field(default_factory=Borders)
    rounding: Rounding = field(default_factory=Rounding)
    headers: Headers = field(default_factory=Headers)

    layouts: Layouts = field(default_factory=Layouts)
    code: CodeBlock = field(default_factory=CodeBlock)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Theme) and self.name == other.name

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the theme."""
        raise NotImplementedError

    def use(self, element: _T) -> _T:
        """Apply the theme to an element.

        Args:
            element (_T): Element to apply the theme to.

        Returns:
            _T: The element with the theme applied.
        """
        element.context["theme"] = self
        return element


@dataclass
class DarkTheme(Theme):
    """The dark theme."""

    name: str = "dark"

    colors: Colors = field(
        default_factory=lambda: Colors(
            primary=ColorRange(["#de946f", "#bd7b5b", "#9e674c"]),
            secondary=Color("#a79272"),
            success=Color("#56A677"),
            info=Color("#518eb5"),
            warning=Color("#d09d3c"),
            danger=Color("#e15c4d"),
            light=ColorRange(["#665949", "#52473a", "#302a22"]),
            dark=ColorRange(["#cfc5b6", "#cab8a5", "#a69685"]),
            white=ColorRange(["#302a22", "#231f1a", "#12100e"]),
            black=ColorRange(["#f2ebe1", "#ebe4da", "#cfc5b6"]),
        ),
    )
    code: CodeBlock = field(
        default_factory=lambda: CodeBlock(
            color=ColorRange(["#ebe4da", "#cfc5b6", "#cab8a5"]),
            background_color=ColorRange(["#52473a", "#302a22", "#231f1a"]),
            line_number_color=Color("#a69685"),
            style=highlight_dark,
        )
    )


@dataclass
class LightTheme(Theme):
    """Light theme."""

    name: str = "light"


_DEFAULT_THEME: Theme = LightTheme()


def get_default_theme() -> Theme:
    """Get the default theme."""
    return _DEFAULT_THEME


def set_default_theme(theme: Theme) -> None:
    """Set the default theme."""
    global _DEFAULT_THEME
    _DEFAULT_THEME = theme



from typing import Annotated, Self, override

from examples import Page, init_db

from ludic.attrs import Attrs
from ludic.catalog.buttons import ButtonPrimary
from ludic.catalog.forms import FieldMeta, Form
from ludic.catalog.headers import H1, H2
from ludic.catalog.layouts import Cluster
from ludic.catalog.quotes import Quote
from ludic.catalog.tables import ColumnMeta, Table, create_rows
from ludic.components import Inline
from ludic.html import span, style
from ludic.web import Endpoint, LudicApp
from ludic.web.parsers import ListParser

db = init_db()
app = LudicApp(debug=True)


class PersonAttrs(Attrs, total=False):
    id: Annotated[str, ColumnMeta(identifier=True)]
    name: Annotated[str, ColumnMeta()]
    email: Annotated[str, ColumnMeta()]
    active: Annotated[
        bool,
        ColumnMeta(kind=FieldMeta(kind="checkbox", label=None)),
    ]


class PeopleAttrs(Attrs):
    people: list[PersonAttrs]


class Toast(Inline):
    id: str = "toast"
    target: str = f"#{id}"
    styles = style.use(
        lambda theme: {
            Toast.target: {
                "background": theme.colors.success,
                "padding": f"{theme.sizes.xxxxs} {theme.sizes.xxxs}",
                "font-size": theme.fonts.size * 0.9,
                "border-radius": "3px",
                "opacity": "0",
                "transition": "opacity 3s ease-out",
            },
            f"{Toast.target}.htmx-settling": {
                "opacity": "100",
            },
        }
    )

    @override
    def render(self) -> span:
        return span(*self.children, id=self.id)


@app.get("/")
async def index() -> Page:
    return Page(
        H1("Bulk Update"),
        Quote(
            "This demo shows how to implement a common pattern where rows are "
            "selected and then bulk updated.",
            source_url="https://htmx.org/examples/bulk-update/",
        ),
        H2("Demo"),
        await PeopleTable.get(),
    )


@app.endpoint("/people/")
class PeopleTable(Endpoint[PeopleAttrs]):
    @classmethod
    async def post(cls, data: ListParser[PersonAttrs]) -> Toast:
        items = {row["id"]: row for row in data.validate()}
        activations = {True: 0, False: 0}

        for person in db.people.values():
            active = items.get(person.id, {}).get("active", False)
            if person.active != active:
                person.active = active
                activations[active] += 1

        return Toast(f"Activated {activations[True]}, deactivated {activations[False]}")

    @classmethod
    async def get(cls) -> Self:
        return cls(people=[person.to_dict() for person in db.people.values()])

    @override
    def render(self) -> Form:
        return Form(
            Table(*create_rows(self.attrs["people"], spec=PersonAttrs)),
            Cluster(
                ButtonPrimary("Bulk Update", type="submit"),
                Toast(),
            ),
            hx_post=self.url_for(PeopleTable),
            hx_target=Toast.target,
            hx_swap="outerHTML settle:3s",
        )



"""This module is designed specifically for use with the mypy plugin."""

from collections.abc import Callable
from typing import TypeVar

from mypy.nodes import ARG_STAR, ARG_STAR2, Argument, TypeInfo, Var
from mypy.plugin import ClassDefContext, Plugin
from mypy.plugins.common import add_method
from mypy.types import (
    AnyType,
    Instance,
    NoneTyp,
    TupleType,
    Type,
    TypeOfAny,
    TypeVarType,
    UnpackType,
)

T = TypeVar("T")
CB = Callable[[T], None] | None

BLACKLISTED_ELEMENTS = {
    "ludic.components.Component",
    "ludic.components.ComponentStrict",
    "ludic.web.endpoints.Endpoint",
}


def is_component_base(info: TypeInfo) -> bool:
    """Check if this is a subclass of a Ludic element."""
    return info.fullname in (
        "ludic.base.Element",
        "ludic.base.ElementStrict",
        "ludic.components.Component",
        "ludic.components.ComponentStrict",
    )


def _find_valid_base(node: TypeInfo) -> Instance | None:
    for base in node.bases:
        if not hasattr(base, "type") or not isinstance(base.type, TypeInfo):
            continue
        if not hasattr(base, "args") or not isinstance(base.args, list | tuple):
            continue
        if is_component_base(base.type) and any(
            not isinstance(arg, TypeVarType) for arg in base.args
        ):
            return base
    return None


def _extract_args_type(ctx: ClassDefContext, base: Instance) -> Type | None:  # noqa
    if not hasattr(base, "type") or not isinstance(base.type, TypeInfo):
        ctx.api.fail(f"Base class {base} does not have a valid TypeInfo.", ctx.cls)
        return None
    if not hasattr(base, "args") or not isinstance(base.args, list | tuple):
        ctx.api.fail(f"Base class {base} does not have valid args.", ctx.cls)
        return None
    if not hasattr(base.type, "name"):
        ctx.api.fail(f"Base class {base} type does not have a name attribute.", ctx.cls)
        return None
    if len(base.args) == 0:
        ctx.api.fail(f"Base class {base} has no type arguments.", ctx.cls)
        return None
    try:
        match base.type.name:
            case "Element" | "Component":
                arg = base.args[0]
                if isinstance(arg, Type):
                    return arg
                ctx.api.fail(
                    f"Base class {base} first arg is not a Type.",
                    ctx.cls,
                )
                return None
            case "ElementStrict" | "ComponentStrict":
                if len(base.args) < 1:
                    ctx.api.fail(
                        f"Base class {base} does not have enough type arguments.",
                        ctx.cls,
                    )
                    return None
                tuple_instance = ctx.api.builtin_type("builtins.tuple")
                return UnpackType(
                    TupleType(
                        list(base.args[:-1]),
                        tuple_instance,
                    )
                )
            case _:
                return None
    except Exception as exc:
        ctx.api.fail(f"Error processing base class {base}: {exc}", ctx.cls)
        return None


def _extract_kwargs_type(ctx: ClassDefContext, base: Instance) -> Type | None:
    if len(base.args) == 0:
        ctx.api.fail(f"Base class {base} has no type arguments for kwargs.", ctx.cls)
        return None
    try:
        last_arg = base.args[-1]
        if isinstance(last_arg, TypeVarType):
            return AnyType(TypeOfAny.special_form)
        return UnpackType(last_arg)
    except Exception as exc:
        ctx.api.fail(
            f"Error processing kwargs type for base class {base}: {exc}",
            ctx.cls,
        )
        return None


class LudicPlugin(Plugin):
    def get_base_class_hook(self, fullname: str) -> CB[ClassDefContext]:
        sym = self.lookup_fully_qualified(fullname)
        if sym and isinstance(sym.node, TypeInfo):
            if is_component_base(sym.node):
                return add_init_hook
        return None


def add_init_hook(ctx: ClassDefContext) -> None:
    """Add a dummy __init__() to a model and record it is generated."""
    node = ctx.cls.info
    if "__init__" in node.names or node.fullname in BLACKLISTED_ELEMENTS:
        return
    base = _find_valid_base(node)
    if base is None:
        return
    args_type = _extract_args_type(ctx, base)
    if args_type is None:
        return
    kwargs_type = _extract_kwargs_type(ctx, base)
    if kwargs_type is None:
        return
    args_var = Var("args", args_type)
    args_arg = Argument(
        variable=args_var,
        type_annotation=args_type,
        initializer=None,
        kind=ARG_STAR,
    )
    kwargs_var = Var("kwargs", kwargs_type)
    kwargs_arg = Argument(
        variable=kwargs_var,
        type_annotation=kwargs_type,
        initializer=None,
        kind=ARG_STAR2,
    )
    add_method(ctx, "__init__", [args_arg, kwargs_arg], NoneTyp())
    ctx.cls.info.metadata.setdefault("ludic", {})["generated_init"] = True


def plugin(version: str) -> type[LudicPlugin]:
    return LudicPlugin



from typing import override

from ludic.attrs import GlobalAttrs, OlAttrs
from ludic.components import Component
from ludic.html import li, ol, ul
from ludic.types import AnyChildren


class Item(Component[AnyChildren, GlobalAttrs]):
    """Simple component simulating an item in a list.

    Example usage:

        Item("Item 1")
    """

    @override
    def render(self) -> li:
        return li(*self.children, **self.attrs)


class List(Component[AnyChildren, GlobalAttrs]):
    """Simple component simulating a list.

    There is basically just an alias for the :class:`ul` element
    without the requirement to pass `li` as children.

    Example usage:

        List("Item 1", "Item 2")
        List(Item("Item 1"), Item("Item 2"))
    """

    template_wrap_in = Item

    @override
    def render(self) -> ul:
        children = (
            child if isinstance(child, Item) else Item(child) for child in self.children
        )
        return ul(*children, **self.attrs_for(ul))


class NumberedList(Component[AnyChildren, OlAttrs]):
    """Simple component simulating a numbered list.

    There is basically just an alias for the :class:`ol` element
    without the requirement to pass `li` as children.

    Example usage:

        NumberedList("Item 1", "Item 2")
        NumberedList(Item("Item 1"), Item("Item 2"))
    """

    template_wrap_in = Item

    @override
    def render(self) -> ol:
        children = (
            child if isinstance(child, Item) else Item(child) for child in self.children
        )
        return ol(*children, **self.attrs_for(ol))



from pygments.style import Style
from pygments.token import (
    Comment,
    Error,
    Generic,
    Keyword,
    Name,
    Number,
    Operator,
    String,
    Whitespace,
)


class LudicLight(Style):
    name = "ludic-light"

    styles = {
        Whitespace: "#bbbbbb",
        Comment: "#888",
        Comment.Preproc: "#579",
        Comment.Special: "bold #cc0000",
        Keyword: "bold #080",
        Keyword.Pseudo: "#038",
        Keyword.Type: "#339",
        Operator: "#333",
        Operator.Word: "bold #000",
        Name.Builtin: "#007020",
        Name.Function: "bold #06B",
        Name.Class: "bold #B06",
        Name.Namespace: "bold #0e84b5",
        Name.Exception: "bold #F00",
        Name.Variable: "#963",
        Name.Variable.Instance: "#33B",
        Name.Variable.Class: "#369",
        Name.Variable.Global: "bold #d70",
        Name.Constant: "bold #036",
        Name.Label: "bold #970",
        Name.Entity: "bold #800",
        Name.Attribute: "#00C",
        Name.Tag: "#070",
        Name.Decorator: "bold #555",
        String: "#BA2121",
        String.Doc: "italic",
        String.Interpol: "bold #A45A77",
        String.Escape: "bold #AA5D1F",
        String.Regex: "#A45A77",
        String.Symbol: "#B8860B",
        String.Symbol: "#19177C",
        String.Other: "#008000",
        Number: "bold #60E",
        Number.Integer: "bold #00D",
        Number.Float: "bold #60E",
        Number.Hex: "bold #058",
        Number.Oct: "bold #40E",
        Generic.Heading: "bold #000080",
        Generic.Subheading: "bold #800080",
        Generic.Deleted: "#A00000",
        Generic.Inserted: "#00A000",
        Generic.Error: "#FF0000",
        Generic.Emph: "italic",
        Generic.Strong: "bold",
        Generic.EmphStrong: "bold italic",
        Generic.Prompt: "bold #c65d09",
        Generic.Output: "#888",
        Generic.Traceback: "#04D",
        Error: "#F00 bg:#FAA",
    }


class LudicDark(Style):
    name = "ludic-dark"

    styles = {
        Whitespace: "#d3c1a6",
        Comment: "#a79272",
        Comment.Preproc: "#518eb5",
        Comment.Special: "bold #bc5f54",
        Keyword: "bold #56a677",
        Keyword.Pseudo: "#518eb5",
        Keyword.Type: "#518eb5",
        Operator: "#d3c1a6",
        Operator.Word: "bold #d3c1a6",
        Name.Builtin: "#56a677",
        Name.Function: "bold #518eb5",
        Name.Class: "bold #e35d6f",
        Name.Namespace: "bold #0e84b5",
        Name.Exception: "bold #bc5f54",
        Name.Variable: "#d09d3c",
        Name.Variable.Instance: "#518eb5",
        Name.Variable.Class: "#518eb5",
        Name.Variable.Global: "bold #d09d3c",
        Name.Constant: "bold #518eb5",
        Name.Label: "bold #d09d3c",
        Name.Entity: "bold #bc5f54",
        Name.Attribute: "#518eb5",
        Name.Tag: "#56a677",
        Name.Decorator: "bold #a79272",
        String: "#bc5f54",
        String.Doc: "italic",
        String.Interpol: "bold #e35d6f",
        String.Escape: "bold #AA5D1F",
        String.Regex: "#e35d6f",
        String.Symbol: "#d09d3c",
        String.Symbol: "#518eb5",
        String.Other: "#56a677",
        Number: "bold #518eb5",
        Number.Integer: "bold #518eb5",
        Number.Float: "bold #518eb5",
        Number.Hex: "bold #518eb5",
        Number.Oct: "bold #518eb5",
        Generic.Heading: "bold #518eb5",
        Generic.Subheading: "bold #800080",
        Generic.Deleted: "#bc5f54",
        Generic.Inserted: "#56a677",
        Generic.Error: "#bc5f54",
        Generic.Emph: "italic",
        Generic.Strong: "bold",
        Generic.EmphStrong: "bold italic",
        Generic.Prompt: "bold #bc5f54",
        Generic.Output: "#d3c1a6",
        Generic.Traceback: "#518eb5",
        Error: "#bc5f54 bg:#e35d6f",
    }



from collections.abc import Callable

from django.http import HttpRequest, HttpResponse


class LudicMiddleware:
    """Ludic middleware for Django.

    With Python 3.14+ t-strings, this middleware is now optional as it no longer
    needs to manage the FormatContext cache. It's kept for backward compatibility.

    Usage:

        # somewhere in django settings.py

        MIDDLEWARES = [
            "...",
            "ludic.contrib.django.LudicMiddleware",
        ]
    """

    def __init__(self, get_response: Callable[[HttpRequest], HttpResponse]) -> None:
        self.get_response = get_response

    def __call__(self, request: HttpRequest) -> HttpResponse:
        response: HttpResponse = self.get_response(request)
        return response



from typing import NotRequired, override

from ludic.attrs import GlobalAttrs
from ludic.components import Component, ComponentStrict
from ludic.html import h2, li, nav, style, ul
from ludic.types import PrimitiveChildren

from .buttons import ButtonLink


class NavItemAttrs(GlobalAttrs):
    to: str
    active: NotRequired[bool]
    active_subsection: NotRequired[bool]


class NavHeader(ComponentStrict[str, GlobalAttrs]):
    """Simple component simulating a navigation header.

    This component is supposed to be used as a child of the :class:`Navigation`
    component.
    """

    classes = ["nav-header"]
    styles = style.use(
        lambda theme: {
            "h2.nav-header": {
                "font-size": theme.headers.h2.size * 0.7,
            }
        }
    )

    @override
    def render(self) -> h2:
        return h2(self.children[0], **self.attrs_for(h2))


class NavItem(Component[PrimitiveChildren, NavItemAttrs]):
    """Simple component simulating a navigation item.

    This component is supposed to be used as a child of the :class:`Navigation`
    component.
    """

    classes = ["nav-item"]
    styles = style.use(
        lambda theme: {
            "li.nav-item.section.active-subsection .btn": {
                "font-weight": "bold",
                "background": "none",
            },
            "li.nav-item.subsection": {
                "padding-inline-start": theme.sizes.m,
            },
        }
    )

    @override
    def render(self) -> li:
        self.attrs.setdefault("classes", [])

        if self.attrs.pop("active_subsection", False):
            self.attrs["classes"].append("active-subsection")

        if self.attrs.pop("active", False):
            self.attrs["classes"].append("active")

        return li(
            ButtonLink(
                self.children[0],
                to=self.attrs["to"],
                external=False,
                classes=["small"] if "subsection" in self.attrs["classes"] else [],
            ),
            **self.attrs_for(li),
        )


class NavSection(ComponentStrict[NavHeader, *tuple[NavItem, ...], GlobalAttrs]):
    """Simple component simulating a navigation section.

    This component is supposed to be used as a child of the :class:`Navigation`
    component.
    """

    @override
    def render(self) -> li:
        self.attrs.setdefault("classes", ["stack", "tiny"])
        return li(
            self.children[0],
            ul(*self.children[1:], classes=["stack", "tiny"]),
            **self.attrs_for(li),
        )


class Navigation(Component[NavItem | NavSection, GlobalAttrs]):
    """Simple component simulating a navigation bar.

    Example usage:

        Navigation(
            NavItem("Home", to="/"),
            NavItem("About", to="/about"),
        )
    """

    classes = ["navigation"]
    styles = {
        "nav.navigation ul": {
            "list-style": "none",
            "padding": "0",
        },
    }

    @override
    def render(self) -> nav:
        return nav(
            ul(*self.children, classes=["stack", "small"]),
            **self.attrs_for(nav),
        )



from typing import override

from examples import DB as Database
from examples import Page, init_db
from fastapi import Depends, FastAPI

from ludic.catalog.forms import Option, SelectField, SelectFieldAttrs
from ludic.catalog.headers import H1, H2
from ludic.catalog.layouts import Stack
from ludic.catalog.loaders import LazyLoader
from ludic.catalog.messages import MessageInfo, Title
from ludic.catalog.quotes import Quote
from ludic.catalog.typography import Code
from ludic.components import Component
from ludic.contrib.fastapi import LudicRoute
from ludic.html import b
from ludic.web import Request
from ludic.web.exceptions import NotFoundError

app = FastAPI()
app.router.route_class = LudicRoute


class CarSelect(Component[str, SelectFieldAttrs]):
    """Ludic element representing car select."""

    @override
    def render(self) -> SelectField:
        return SelectField(
            *[Option(child, value=child.lower()) for child in self.children],
            label=self.attrs.pop("label", "Car Manufacturer"),
            name="manufacturer",
            **self.attrs,
        )


class CarModelsSelect(Component[str, SelectFieldAttrs]):
    """Ludic element representing car models select."""

    @override
    def render(self) -> SelectField:
        return SelectField(
            *[Option(child, value=child.lower()) for child in self.children],
            label=self.attrs.pop("label", "Car Model"),
            id="models",
            **self.attrs,
        )


@app.get("/")
async def index(request: Request) -> Page:
    return Page(
        H1("Cascading Select"),
        MessageInfo(
            Title("FastAPI Example"),
            t"This example uses {b('FastAPI')} as backend Web Framework.",
        ),
        Quote(
            t"In this example we show how to make the values in one {Code('select')} "
            t"depend on the value selected in another {Code('select')}.",
            source_url="https://htmx.org/examples/value-select/",
        ),
        H2("Demo"),
        LazyLoader(load_url=request.url_for(cars)),
    )


@app.get("/cars/")
def cars(request: Request, db: Database = Depends(init_db)) -> Stack:
    return Stack(
        CarSelect(
            *db.find_all_cars_names(),
            hx_get=request.url_for(models),
            hx_target="#models",
        ),
        CarModelsSelect(*db.find_first_car().models),
    )


@app.get("/models/")
def models(
    manufacturer: str | None = None, db: Database = Depends(init_db)
) -> CarModelsSelect:
    if car := db.find_car_by_name(manufacturer):
        return CarModelsSelect(*car.models, label=None)  # type: ignore
    else:
        raise NotFoundError("Car could not be found")



import html
import inspect
import itertools
from collections.abc import Mapping
from functools import lru_cache
from string.templatelib import Interpolation
from string.templatelib import Template as Template
from typing import Any, TypeVar, get_type_hints

T = TypeVar("T")


@lru_cache
def _load_attrs_aliases() -> Mapping[str, str]:
    from ludic import attrs

    result = {}
    for name, cls in inspect.getmembers(attrs, inspect.isclass):
        if not name.endswith("Attrs"):
            continue

        hints = get_type_hints(cls, include_extras=True)
        for key, value in hints.items():
            if metadata := getattr(value, "__metadata__", None):
                for meta in metadata:
                    if isinstance(meta, attrs.Alias):
                        result[key] = str(meta)

    return result


def format_attr_value(key: str, value: Any, is_html: bool = False) -> str:
    """Format an HTML attribute with the given key and value.

    Args:
        key (str): The key of the attribute.
        value (Any): The value of the attribute, can be a string or a dictionary.
    Returns:
        str: The formatted HTML attribute.
    """

    def _html_escape(value: Any) -> str:
        if (
            is_html
            and value
            and isinstance(value, str)
            and getattr(value, "escape", True)
        ):
            return html.escape(value, False)  # type: ignore
        return str(value)

    if isinstance(value, dict):
        formatted_value = ";".join(
            f"{dict_key}:{_html_escape(dict_value)}"
            for dict_key, dict_value in value.items()
        )
    elif isinstance(value, list):
        formatted_value = " ".join(map(_html_escape, value))
    elif isinstance(value, bool):
        if is_html and not key.startswith("hx"):
            formatted_value = html.escape(key, False) if value else ""
        else:
            formatted_value = "true" if value else "false"
    else:
        formatted_value = _html_escape(value)

    return formatted_value


def extract_dataset_attrs(attrs: Mapping[str, Any]) -> Mapping[str, Any]:
    """Converts "dataset" attributes to "data-*" attributes names.

    Args:
        attrs (Mapping[str, Any]): A dictionary containing HTML attributes.

    Returns:
        Mapping[str, Any]: A dictionary containing extracted data-* attributes.
    """
    extracted_attrs = {}
    if (dataset_attrs := attrs.get("dataset")) and isinstance(dataset_attrs, dict):
        for key, value in dataset_attrs.items():
            extracted_attrs[f"data-{key}"] = value
    return extracted_attrs


def format_attrs(attrs: Mapping[str, Any], is_html: bool = False) -> dict[str, Any]:
    """Format the given attributes.

        attrs = {"name": "John", "class_": "person", "is_adult": True}

    The result will be:

        >>> format_attrs(attrs)
        >>> {"name": "John", "class": "person"}

    Args:
        attrs (Mapping[str, Any]): The attributes to format.

    Returns:
        Mapping[str, Any]: The formatted attributes.
    """
    aliases = _load_attrs_aliases()
    result: dict[str, str] = {}
    dataset_attrs = extract_dataset_attrs(attrs)

    for key, value in itertools.chain(attrs.items(), dataset_attrs.items()):
        if key == "dataset":
            continue
        if formatted_value := format_attr_value(key, value, is_html=is_html):
            if key in aliases:
                alias = aliases[key]
            else:
                alias = key.strip("_").replace("_", "-")

            if alias in result:
                result[alias] += " " + formatted_value
            else:
                result[alias] = formatted_value
    return result


def format_element(child: Any) -> str:
    """Default HTML formatter.

    Args:
        child (AnyChild): The HTML element or text to format.
    """
    if isinstance(child, str) and getattr(child, "escape", True):
        return html.escape(child, False)
    elif hasattr(child, "to_html"):
        return child.to_html()  # type: ignore
    else:
        return str(child)


def process_template(template: Any, wrap_in: type | None = None) -> tuple[Any, ...]:
    """Process a t-string template into a tuple of children.

    This function processes Python 3.14 t-string Template objects, extracting
    both static string parts and dynamic interpolated values. This replaces the
    old FormatContext system.

    Example usage:

        dom = div(t"test {b('foo')} {i('bar')}")
        # Internally processes to: div("test ", b('foo'), " ", i('bar'))

    Args:
        template: The Template object from a t-string literal.
        wrap_in: Optional element type to wrap the processed parts in.

    Returns:
        tuple[Any, ...]: A tuple of children elements and strings.
    """
    # Template objects are iterable, yielding str or Interpolation objects
    parts: list[Any] = []
    for part in template:
        if isinstance(part, str):
            # Static string part
            if part:  # Only add non-empty strings
                parts.append(part)
        elif isinstance(part, Interpolation):
            # Dynamic interpolated value
            value = part.value
            parts.append(value)
        else:
            # Fallback for unexpected types
            parts.append(part)

    if wrap_in is not None:
        return (wrap_in(*parts),)

    return tuple(parts)



from abc import ABCMeta
from collections.abc import Iterator, Mapping, Sequence
from typing import Any, ClassVar

from .format import Template, format_attrs, format_element, process_template


class BaseElement(metaclass=ABCMeta):
    html_header: ClassVar[str | None] = None
    html_name: ClassVar[str | None] = None
    void_element: ClassVar[bool] = False

    # Support for wrapping processed template strings in a specific element
    template_wrap_in: ClassVar[type[BaseElement] | None] = None

    children: Sequence[Any]
    attrs: Mapping[str, Any]
    context: dict[str, Any]

    def __init__(self, *children: Any, **attrs: Any) -> None:
        self.context = {}
        # Process t-string templates in children
        processed_children: list[Any] = []
        for child in children:
            if isinstance(child, Template):
                # Process t-string template
                processed_children.extend(
                    process_template(child, wrap_in=self.template_wrap_in)
                )
            else:
                processed_children.append(child)
        self.children = tuple(processed_children)
        self.attrs = attrs

    def __str__(self) -> str:
        return self.to_html()

    def __bytes__(self) -> bytes:
        return self.to_html().encode("utf-8")

    def __len__(self) -> int:
        return len(self.children)

    def __iter__(self) -> Iterator[Any]:
        return iter(self.children)

    def __repr__(self) -> str:
        return self.to_string(pretty=False)

    def __eq__(self, other: Any) -> bool:
        return (
            type(self) is type(other)
            and self.children == other.children
            and self.attrs == other.attrs
        )

    def _format_attributes(self, is_html: bool = False) -> str:
        attrs: dict[str, Any] = format_attrs(self.attrs, is_html=is_html)
        return " ".join(
            f'{key}="{value}"' if '"' not in value else f"{key}='{value}'"
            for key, value in attrs.items()
        )

    def _format_children(self) -> str:
        formatted = ""
        for child in self.children:
            if self.context and isinstance(child, BaseElement):
                child.context.update(self.context)
            formatted += format_element(child)
        return formatted

    @property
    def aliased_attrs(self) -> dict[str, Any]:
        """Attributes as a dict with keys renamed to their aliases."""
        return format_attrs(self.attrs)

    @property
    def text(self) -> str:
        """Get the text content of the element."""
        return "".join(
            child.text if isinstance(child, BaseElement) else str(child)
            for child in self.children
        )

    def is_simple(self) -> bool:
        """Check if the element is simple (i.e. contains only one primitive type)."""
        return len(self) == 1 and isinstance(self.children[0], str | int | float | bool)

    def has_attributes(self) -> bool:
        """Check if the element has any attributes."""
        return bool(self.attrs)

    def to_string(self, pretty: bool = True, _level: int = 0) -> str:
        """Convert the element tree to a string representation.

        Args:
            pretty (bool, optional): Whether to indent the string. Defaults to True.

        Returns:
            str: The string representation of the element tree.
        """
        indent = "  " * _level if pretty else ""
        name = self.__class__.__name__
        element = f"<{name}"

        if self.has_attributes():
            element += f" {self._format_attributes()}"

        if self.children:
            prefix, sep, suffix = "", "", ""
            if pretty and (not self.is_simple() or self.has_attributes()):
                prefix, sep, suffix = f"\n{indent}  ", f"\n{indent}  ", f"\n{indent}"

            children_str = sep.join(
                child.to_string(pretty=pretty, _level=_level + 1)
                if isinstance(child, BaseElement)
                else str(child)
                for child in self.children
            )

            element += f">{prefix}{children_str}{suffix}</{name}>"
        else:
            element += " />"

        return element

    def to_html(self) -> str:
        """Convert an element tree to an HTML string."""
        element_tag = f"{self.html_header}\n" if self.html_header else ""
        children_str = self._format_children() if self.children else ""

        element_tag += f"<{self.html_name}"
        if self.has_attributes():
            attributes_str = self._format_attributes(is_html=True)
            element_tag += f" {attributes_str}"

        if not self.void_element:
            element_tag += f">{children_str}</{self.html_name}>"
        else:
            element_tag += ">"

        return element_tag



from collections.abc import Iterable, Mapping
from typing import Never, TypeAlias, TypedDict, TypeVar, TypeVarTuple

from .attrs import Attrs, NoAttrs, URLType
from .base import BaseElement
from .styles import CSSProperties, GlobalStyles


class Safe(str):
    """Marker for a string that is safe to use as is without HTML escaping.

    That means the content of the string is not escaped when rendered.

    Usage:

        >>> div(Safe("Hello <b>World!</b>")).to_html()
        '<div>Hello <b>World!</b></div>'
    """

    escape = False


class JavaScript(Safe):
    """Marker for a JavaScript string.

    The content of this string is not escaped when rendered.

    Usage:

        js = JavaScript("alert('Hello World!')")
    """


JSONType = (
    Mapping[str, "JSONType"] | Iterable["JSONType"] | str | int | float | bool | None
)
Headers = Mapping[str, JSONType]
HXHeaders = TypedDict(
    "HXHeaders",
    {
        "HX-Location": JSONType,
        "HX-Push-Url": URLType | bool,
        "HX-Redirect": str,
        "HX-Refresh": bool,
        "HX-Replace-Url": URLType,
        "HX-Reswap": str,
        "HX-Retarget": str,
        "HX-Reselect": str,
        "HX-Trigger": JSONType,
        "HX-Trigger-After-Settle": JSONType,
        "HX-Trigger-After-Swap": JSONType,
    },
    total=False,
)

NoChildren: TypeAlias = Never
"""Type alias for elements that are not allowed to have children."""

PrimitiveChildren: TypeAlias = str | bool | int | float
"""Type alias for elements that are allowed to have only primitive children.

Primitive children are ``str``, ``bool``, ``int`` and ``float``.
"""

ComplexChildren: TypeAlias = BaseElement
"""Type alias for elements that are allowed to have only non-primitive children."""

AnyChildren: TypeAlias = PrimitiveChildren | ComplexChildren | Safe
"""Type alias for elements that are allowed to have any children."""

TChildren = TypeVar("TChildren", bound=AnyChildren, covariant=True)
"""Type variable for elements representing type of children (the type of *args).

See also: :class:`ludic.types.Component`.
"""

TChildrenArgs = TypeVarTuple("TChildrenArgs")
"""Type variable for strict elements representing type of children (the type of *args).

See also: :class:`ludic.types.ComponentStrict`.
"""

TAttrs = TypeVar("TAttrs", bound=Attrs | NoAttrs, covariant=True)
"""Type variable for elements representing type of attributes (the type of **kwargs)."""


__all__ = (
    "AnyChildren",
    "ComplexChildren",
    "CSSProperties",
    "GlobalStyles",
    "Headers",
    "HXHeaders",
    "URLType",
    "JavaScript",
    "NoChildren",
    "PrimitiveChildren",
    "Safe",
    "TAttrs",
    "TChildren",
    "TChildrenArgs",
)



from collections.abc import Callable


def attr_to_camel(name: str) -> str:
    """Convert an attribute name to camel case.

    Args:
        name (str): The attribute name.

    Returns:
        str: The attribute name in camel case.
    """
    return " ".join(value.title() for value in name.split("_"))


def text_to_kebab(name: str) -> str:
    """Convert a text to kebab case.

    Args:
        name (str): The text.

    Returns:
        str: The text in kebab case.
    """
    return "-".join(value.lower() for value in name.split())


def remove_whitespaces(text: str) -> str:
    """Remove leading whitespaces from a text.

    Args:
        text (str): The text.

    Returns:
        str: The text without leading whitespaces.
    """
    if not text:
        return text

    by_line = text.splitlines()
    min_whitespaces = min(len(line) - len(line.lstrip()) for line in by_line if line)
    return "\n".join(line[min_whitespaces:].rstrip() for line in by_line).strip()


def add_line_numbers(text: str, apply_fun: Callable[[str], str] = str) -> str:
    """Add line number at the beginning of each line.

    Args:
        text (str): Content to append line number.
        apply_fun (Callable[[int], str]): Call this function on each number.

    Returns:
        str: Text with number appended to each line.
    """
    lines = text.splitlines()
    max_digits = len(str(len(lines)))

    return "\n".join(
        f"{apply_fun(f'{number + 1:>{max_digits}}   ')}{line}"
        for number, line in enumerate(lines)
    )



from typing import override

from ludic.attrs import Attrs
from ludic.components import Component
from ludic.html import div
from ludic.types import AnyChildren


class ClassesComponentAttrs(Attrs):
    class_: str


class ClassesComponent(Component[AnyChildren, ClassesComponentAttrs]):
    classes = ["class-a"]

    @override
    def render(self) -> div:
        return div(*self.children, classes=["class-b"], **self.attrs)


def test_component_classes() -> None:
    assert ClassesComponent(
        div("content", class_="class-f"), class_="class-c class-d"
    ).to_html() == (
        '<div class="class-b class-a class-c class-d">'
            '<div class="class-f">content</div>'
        '</div>'
    )  # fmt: skip



from typing import ClassVar, Generic, Unpack

from .attrs import NoAttrs
from .base import BaseElement
from .types import TAttrs, TChildren, TChildrenArgs


class Element(Generic[TChildren, TAttrs], BaseElement):
    """Base class for Ludic elements.

    Args:
        *children (TChild): The children of the element.
        **attrs (Unpack[TAttrs]): The attributes of the element.
    """

    children: tuple[TChildren, ...]
    attrs: TAttrs

    formatter_wrap_in: ClassVar[type[BaseElement] | None] = None

    def __init__(
        self,
        *children: TChildren,
        # FIXME: https://github.com/python/typing/issues/1399
        **attrs: Unpack[TAttrs],
    ) -> None:
        super().__init__(*children, **attrs)


class ElementStrict(Generic[*TChildrenArgs, TAttrs], BaseElement):
    """Base class for strict elements (elements with concrete types of children).

    Args:
        *children (*TChildTuple): The children of the element.
        **attrs (Unpack[TAttrs]): The attributes of the element.
    """

    children: tuple[*TChildrenArgs]
    attrs: TAttrs

    formatter_wrap_in: ClassVar[type[BaseElement] | None] = None

    def __init__(
        self,
        *children: *TChildrenArgs,
        # FIXME: https://github.com/python/typing/issues/1399
        **attrs: Unpack[TAttrs],
    ) -> None:
        super().__init__(*children, **attrs)


class Blank(Element[TChildren, NoAttrs]):
    """Element representing no element at all, just children.

    The purpose of this element is to be able to return only children
    when rendering a component.
    """

    def __init__(self, *children: TChildren) -> None:
        super().__init__(*children)

    def to_html(self) -> str:
        return "".join(map(str, self.children))



import asyncio
import inspect
from collections.abc import Generator
from typing import Any

import pytest
from starlette._utils import AwaitableOrContextManager, AwaitableOrContextManagerWrapper
from starlette.applications import Starlette
from starlette.datastructures import FormData, Headers, QueryParams
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Route
from starlette.testclient import TestClient
from starlette.websockets import WebSocket

from ludic.base import BaseElement
from ludic.web.parsers import BaseParser
from ludic.web.responses import (
    LudicResponse,
    extract_from_request,
    extract_response_status_headers,
    prepare_response,
)


class DummyElement(BaseElement):
    def to_html(self) -> str:
        return "<p>Hello World</p>"


class DummyParser(BaseParser[Any]):
    def __init__(self, form: FormData) -> None:
        self.form = form


class DummyRequest(Request):
    path_params: dict[str, str] = {}
    query_params: QueryParams = QueryParams({"foo": "bar"})
    headers: Headers = Headers({"x": "y"})

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    def form(self, *args: Any, **kwargs: Any) -> AwaitableOrContextManager[FormData]:
        class DummyForm(FormData):
            async def __aenter__(self) -> Any:
                return self

            async def __aexit__(self, *a: Any) -> None:
                pass

            def __await__(self) -> Generator[None, None, FormData]:
                yield None
                return self

        return AwaitableOrContextManagerWrapper(DummyForm())


class DummyWebSocket(WebSocket):
    path_params: dict[str, str] = {}
    query_params: QueryParams = QueryParams({"foo": "bar"})
    headers: Headers = Headers({"x": "y"})

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass


def plain_handler() -> str:
    return "Hello"


def tuple_handler() -> tuple[str, int]:
    return "Hello", 201


def triple_handler() -> tuple[str, int, dict[str, str]]:
    return "Hello", 202, {"x-header": "yes"}


def invalid_tuple() -> tuple[str, int, dict[str, str], str]:
    return "Invalid", 201, {"x": "a"}, "extra"


def test_extract_response_status_headers_two_elements() -> None:
    body, code, headers = extract_response_status_headers(("hello", 404))
    assert isinstance(body, str)
    assert body == "hello"
    assert code == 404
    assert headers is None


def test_extract_response_status_headers_three_elements() -> None:
    _, _, headers = extract_response_status_headers(("hello", 200, {"x": "1"}))
    assert headers is not None and headers["x"] == "1"


def test_extract_response_status_headers_invalid_tuple() -> None:
    with pytest.raises(ValueError):
        extract_response_status_headers(("x", 1, {}, "extra"))


def test_ludic_response_render() -> None:
    response = LudicResponse(DummyElement())
    assert b"<p>Hello World</p>" in response.body


async def plain_view(request: Request) -> Response:
    return await prepare_response(lambda: "Hello plain", request)


async def tuple_view(request: Request) -> Response:
    return await prepare_response(lambda: ("Hi", 203), request)


async def element_view(request: Request) -> Response:
    return await prepare_response(lambda: DummyElement(), request)


app = Starlette(
    routes=[
        Route("/plain", plain_view),
        Route("/tuple", tuple_view),
        Route("/element", element_view),
    ]
)

client = TestClient(app)


def test_prepare_response_plaintext() -> None:
    response = client.get("/plain")
    assert response.status_code == 200
    assert response.text == "Hello plain"


def test_prepare_response_tuple() -> None:
    response = client.get("/tuple")
    assert response.status_code == 203
    assert response.text == "Hi"


def test_prepare_response_ludic_element() -> None:
    response = client.get("/element")
    assert response.status_code == 200
    assert "<p>Hello World</p>" in response.text


def test_extract_response_status_headers_invalid_length() -> None:
    with pytest.raises(ValueError):
        extract_response_status_headers(("a",))
    with pytest.raises(ValueError):
        extract_response_status_headers(("a", 1, {}, "extra"))


def test_prepare_response_invalid_type() -> None:
    async def handler() -> Any:
        return object()

    async def route_handler(req: Request) -> Response:
        return await prepare_response(handler, req)

    app = Starlette(routes=[Route("/invalid", route_handler)])
    client = TestClient(app)
    with pytest.raises(ValueError):
        client.get("/invalid")


def test_prepare_response_none() -> None:
    async def handler() -> None:
        return None

    async def route_handler(req: Request) -> Response:
        return await prepare_response(handler, req)

    app = Starlette(routes=[Route("/none", route_handler)])
    client = TestClient(app)
    resp = client.get("/none")
    assert resp.status_code == 204


def test_extract_from_request_issubclass_typeerror() -> None:
    def handler(x: int) -> int:
        return x

    param = list(inspect.signature(handler).parameters.values())[0]
    param = param.replace(annotation=123)
    result = asyncio.run(extract_from_request(lambda x: x, DummyRequest()))

    assert isinstance(result, dict)


def test_extract_from_request_invalid_union() -> None:
    def handler(x: int | str) -> int | str:
        return x

    with pytest.raises(TypeError):
        asyncio.run(extract_from_request(handler, DummyRequest()))


def test_extract_from_request_empty_annotation() -> None:
    def handler(x: Any) -> Any:
        return x

    result = asyncio.run(extract_from_request(handler, DummyRequest()))
    assert isinstance(result, dict)


def test_extract_from_request_unknown_type() -> None:
    class Unknown:
        pass

    def handler(x: Unknown) -> Unknown:
        return x

    result = asyncio.run(extract_from_request(handler, DummyRequest()))
    assert isinstance(result, dict)


def test_extract_from_request_all_types() -> None:
    def handler(
        ws: WebSocket,
        parser: DummyParser,
        form: FormData,
        req: Request,
        qp: QueryParams,
        hdr: Headers,
    ) -> None:
        return None

    request = DummyRequest()
    request.__class__.__name__ = "WebSocket"
    result = asyncio.run(extract_from_request(handler, request))
    assert set(result.keys()) == {"form", "req", "qp", "hdr"}

    websocket_request = DummyWebSocket()
    websocket_result = asyncio.run(extract_from_request(handler, websocket_request))
    assert set(websocket_result.keys()) == {"ws"}


def test_prepare_response_handler_exception() -> None:
    def handler() -> None:
        raise RuntimeError("fail")

    async def route_handler(req: Request) -> Response:
        return await prepare_response(handler, req)

    app = Starlette(routes=[Route("/fail", route_handler)])
    client = TestClient(app)

    with pytest.raises(RuntimeError):
        client.get("/fail")



from typing import override

from ludic.attrs import Attrs
from ludic.base import BaseElement
from ludic.components import ComponentStrict
from ludic.html import a, blockquote, div, footer, p, style
from ludic.types import AnyChildren


class QuoteAttrs(Attrs, total=False):
    source_url: str
    source_text: str


class Quote(ComponentStrict[str, QuoteAttrs]):
    """Simple component rendering as the HTML ``blockquote`` element."""

    classes = ["quote"]
    styles = style.use(
        lambda theme: {
            ".quote": {
                "blockquote": {
                    "background-color": theme.colors.light,
                    "border-left": (
                        f"{theme.borders.thick} solid {theme.colors.light.darken(1)}"
                    ),
                    "margin-bottom": theme.sizes.xxs,
                    "padding": f"{theme.sizes.l} {theme.sizes.m}",
                },
                "blockquote p": {
                    "font-size": theme.fonts.size,
                },
                "blockquote code": {
                    "background-color": theme.colors.white,
                },
                "footer": {
                    "font-size": theme.fonts.size * 0.9,
                    "color": theme.colors.dark.lighten(1),
                },
                "footer a": {
                    "text-decoration": "none",
                    "color": theme.colors.dark.darken(1),
                },
                "footer a:hover": {
                    "text-decoration": "underline",
                },
            }
        }
    )

    @override
    def render(self) -> div:
        p_children: list[BaseElement] = []
        current_children: list[AnyChildren] = []

        for child in self.children:
            if isinstance(child, str) and "\n\n" in child:
                p_children.append(p(*current_children))
                current_children = []
            else:
                current_children.append(child)

        children: list[BaseElement] = [blockquote(*p_children, p(*current_children))]
        if source_url := self.attrs.get("source_url"):
            source_text = self.attrs.get("source_text", "Source: ")
            children.append(footer(source_text, a(source_url, href=source_url)))
        return div(*children)



from collections.abc import Callable, Iterator
from typing import Generic, Self, Unpack

from .attrs import (
    AreaAttrs,
    AudioAttrs,
    BaseAttrs,
    BlockquoteAttrs,
    ButtonAttrs,
    CanvasAttrs,
    CircleAttrs,
    ColAttrs,
    DataAttrs,
    DelAttrs,
    DetailsAttrs,
    DialogAttrs,
    EmbedAttrs,
    FieldsetAttrs,
    FormAttrs,
    GlobalAttrs,
    HeadLinkAttrs,
    HtmlAttrs,
    HtmlTagAttrs,
    HyperlinkAttrs,
    IframeAttrs,
    ImgAttrs,
    InputAttrs,
    InsAttrs,
    LabelAttrs,
    LiAttrs,
    LineAttrs,
    MapAttrs,
    MetaAttrs,
    MeterAttrs,
    NoAttrs,
    ObjectAttrs,
    OlAttrs,
    OptgroupAttrs,
    OptionAttrs,
    OutputAttrs,
    ParamAttrs,
    PathAttrs,
    PolylineAttrs,
    ProgressAttrs,
    QAttrs,
    ScriptAttrs,
    SelectAttrs,
    SourceAttrs,
    StyleAttrs,
    SvgAttrs,
    TdAttrs,
    TextAreaAttrs,
    ThAttrs,
    TimeAttrs,
    TrackAttrs,
    VideoAttrs,
)
from .base import BaseElement
from .elements import Element, ElementStrict
from .styles import (
    format_styles,
    from_components,
    from_loaded,
)
from .styles.types import TTheme
from .types import (
    AnyChildren,
    ComplexChildren,
    CSSProperties,
    GlobalStyles,
    NoChildren,
    PrimitiveChildren,
)


class div(Element[AnyChildren, GlobalAttrs]):
    html_name = "div"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class span(Element[AnyChildren, GlobalAttrs]):
    html_name = "span"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class main(Element[AnyChildren, GlobalAttrs]):
    html_name = "main"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class p(Element[AnyChildren, GlobalAttrs]):
    html_name = "p"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class a(Element[AnyChildren, HyperlinkAttrs]):
    html_name = "a"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[HyperlinkAttrs]) -> None:
        super().__init__(*children, **attrs)


class br(Element[NoChildren, GlobalAttrs]):
    void_element = True
    html_name = "br"

    def __init__(self, *children: NoChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class button(Element[AnyChildren, ButtonAttrs]):
    html_name = "button"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[ButtonAttrs]) -> None:
        super().__init__(*children, **attrs)


class label(Element[AnyChildren, LabelAttrs]):
    html_name = "label"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[LabelAttrs]) -> None:
        super().__init__(*children, **attrs)


class td(Element[AnyChildren, TdAttrs]):
    html_name = "td"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[TdAttrs]) -> None:
        super().__init__(*children, **attrs)


class th(Element[AnyChildren, ThAttrs]):
    html_name = "th"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[ThAttrs]) -> None:
        super().__init__(*children, **attrs)


class tr(Element[ComplexChildren, GlobalAttrs]):
    html_name = "tr"

    def __init__(
        self, *children: ComplexChildren, **attrs: Unpack[GlobalAttrs]
    ) -> None:
        super().__init__(*children, **attrs)


class thead(Element[ComplexChildren, GlobalAttrs]):
    html_name = "thead"

    def __init__(
        self, *children: ComplexChildren, **attrs: Unpack[GlobalAttrs]
    ) -> None:
        super().__init__(*children, **attrs)


class tbody(Element[ComplexChildren, GlobalAttrs]):
    html_name = "tbody"

    def __init__(
        self, *children: ComplexChildren, **attrs: Unpack[GlobalAttrs]
    ) -> None:
        super().__init__(*children, **attrs)


class tfoot(Element[ComplexChildren, GlobalAttrs]):
    html_name = "tfoot"

    def __init__(
        self, *children: ComplexChildren, **attrs: Unpack[GlobalAttrs]
    ) -> None:
        super().__init__(*children, **attrs)


class table(Element[ComplexChildren, GlobalAttrs]):
    html_name = "table"

    def __init__(
        self, *children: ComplexChildren, **attrs: Unpack[GlobalAttrs]
    ) -> None:
        super().__init__(*children, **attrs)


class li(Element[AnyChildren, LiAttrs]):
    html_name = "li"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[LiAttrs]) -> None:
        super().__init__(*children, **attrs)


class ul(Element[ComplexChildren, GlobalAttrs]):
    html_name = "ul"

    def __init__(
        self, *children: ComplexChildren, **attrs: Unpack[GlobalAttrs]
    ) -> None:
        super().__init__(*children, **attrs)


class ol(Element[ComplexChildren, OlAttrs]):
    html_name = "ol"

    def __init__(self, *children: ComplexChildren, **attrs: Unpack[OlAttrs]) -> None:
        super().__init__(*children, **attrs)


class dt(Element[AnyChildren, GlobalAttrs]):
    html_name = "dt"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class dd(Element[AnyChildren, GlobalAttrs]):
    html_name = "dd"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class dl(Element[ComplexChildren, GlobalAttrs]):
    html_name = "dl"

    def __init__(
        self, *children: ComplexChildren, **attrs: Unpack[GlobalAttrs]
    ) -> None:
        super().__init__(*children, **attrs)


class section(Element[AnyChildren, GlobalAttrs]):
    html_name = "section"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class input(Element[NoChildren, InputAttrs]):
    void_element = True
    html_name = "input"

    def __init__(self, *children: NoChildren, **attrs: Unpack[InputAttrs]) -> None:
        super().__init__(*children, **attrs)


class output(Element[NoChildren, OutputAttrs]):
    html_name = "output"

    def __init__(self, *children: NoChildren, **attrs: Unpack[OutputAttrs]) -> None:
        super().__init__(*children, **attrs)


class legend(Element[PrimitiveChildren, GlobalAttrs]):
    html_name = "legend"

    def __init__(
        self, *children: PrimitiveChildren, **attrs: Unpack[GlobalAttrs]
    ) -> None:
        super().__init__(*children, **attrs)


class option(Element[PrimitiveChildren, OptionAttrs]):
    html_name = "option"

    def __init__(
        self, *children: PrimitiveChildren, **attrs: Unpack[OptionAttrs]
    ) -> None:
        super().__init__(*children, **attrs)


class optgroup(Element[AnyChildren, OptgroupAttrs]):
    html_name = "optgroup"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[OptgroupAttrs]) -> None:
        super().__init__(*children, **attrs)


class select(Element[AnyChildren, SelectAttrs]):
    html_name = "select"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[SelectAttrs]) -> None:
        super().__init__(*children, **attrs)


class textarea(Element[PrimitiveChildren, TextAreaAttrs]):
    html_name = "textarea"

    def __init__(
        self, *children: PrimitiveChildren, **attrs: Unpack[TextAreaAttrs]
    ) -> None:
        super().__init__(*children, **attrs)


class fieldset(Element[AnyChildren, FieldsetAttrs]):
    html_name = "fieldset"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[FieldsetAttrs]) -> None:
        super().__init__(*children, **attrs)


class form(Element[AnyChildren, FormAttrs]):
    html_name = "form"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[FormAttrs]) -> None:
        super().__init__(*children, **attrs)


class img(Element[NoChildren, ImgAttrs]):
    void_element = True
    html_name = "img"

    def __init__(self, *children: NoChildren, **attrs: Unpack[ImgAttrs]) -> None:
        super().__init__(*children, **attrs)


class svg(Element[AnyChildren, SvgAttrs]):
    html_name = "svg"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[SvgAttrs]) -> None:
        super().__init__(*children, **attrs)


class circle(Element[AnyChildren, CircleAttrs]):
    html_name = "circle"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[CircleAttrs]) -> None:
        super().__init__(*children, **attrs)


class line(Element[AnyChildren, LineAttrs]):
    html_name = "line"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[LineAttrs]) -> None:
        super().__init__(*children, **attrs)


class path(Element[AnyChildren, PathAttrs]):
    html_name = "path"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[PathAttrs]) -> None:
        super().__init__(*children, **attrs)


class polyline(Element[AnyChildren, PolylineAttrs]):
    html_name = "polyline"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[PolylineAttrs]) -> None:
        super().__init__(*children, **attrs)


class b(Element[AnyChildren, GlobalAttrs]):
    html_name = "b"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class i(Element[AnyChildren, GlobalAttrs]):
    html_name = "i"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class s(Element[AnyChildren, GlobalAttrs]):
    html_name = "s"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class u(Element[AnyChildren, GlobalAttrs]):
    html_name = "u"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class strong(Element[AnyChildren, GlobalAttrs]):
    html_name = "strong"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class em(Element[AnyChildren, GlobalAttrs]):
    html_name = "em"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class mark(Element[AnyChildren, GlobalAttrs]):
    html_name = "mark"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class del_(Element[AnyChildren, DelAttrs]):
    html_name = "del"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[DelAttrs]) -> None:
        super().__init__(*children, **attrs)


class ins(Element[AnyChildren, InsAttrs]):
    html_name = "ins"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[InsAttrs]) -> None:
        super().__init__(*children, **attrs)


class header(Element[AnyChildren, GlobalAttrs]):
    html_name = "header"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class big(Element[AnyChildren, GlobalAttrs]):
    html_name = "big"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class small(Element[AnyChildren, GlobalAttrs]):
    html_name = "small"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class code(Element[AnyChildren, GlobalAttrs]):
    html_name = "code"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class pre(Element[AnyChildren, GlobalAttrs]):
    html_name = "pre"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class cite(Element[AnyChildren, GlobalAttrs]):
    html_name = "cite"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class blockquote(Element[AnyChildren, BlockquoteAttrs]):
    html_name = "blockquote"

    def __init__(
        self, *children: AnyChildren, **attrs: Unpack[BlockquoteAttrs]
    ) -> None:
        super().__init__(*children, **attrs)


class abbr(Element[AnyChildren, GlobalAttrs]):
    html_name = "abbr"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class h1(Element[AnyChildren, GlobalAttrs]):
    html_name = "h1"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class h2(Element[AnyChildren, GlobalAttrs]):
    html_name = "h2"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class h3(Element[AnyChildren, GlobalAttrs]):
    html_name = "h3"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class h4(Element[AnyChildren, GlobalAttrs]):
    html_name = "h4"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class h5(Element[AnyChildren, GlobalAttrs]):
    html_name = "h5"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class h6(Element[AnyChildren, GlobalAttrs]):
    html_name = "h6"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class title(Element[PrimitiveChildren, NoAttrs]):
    html_name = "title"

    def __init__(self, *children: PrimitiveChildren, **attrs: Unpack[NoAttrs]) -> None:
        super().__init__(*children, **attrs)


class link(Element[NoChildren, HeadLinkAttrs]):
    void_element = True
    html_name = "link"

    def __init__(self, *children: NoChildren, **attrs: Unpack[HeadLinkAttrs]) -> None:
        super().__init__(*children, **attrs)


class style(Generic[TTheme], BaseElement, GlobalStyles):
    html_name = "style"

    children: tuple[GlobalStyles | Callable[[TTheme], GlobalStyles] | str]
    attrs: StyleAttrs

    def __init__(
        self,
        styles: GlobalStyles | Callable[[TTheme], GlobalStyles] | str,
        theme: TTheme | None = None,
        **attrs: Unpack[StyleAttrs],
    ) -> None:
        super().__init__(styles, **attrs)

        if theme:
            self.context["theme"] = theme

    @classmethod
    def use(cls, styles: GlobalStyles | Callable[[TTheme], GlobalStyles]) -> Self:
        return cls(styles)

    @classmethod
    def from_components(
        cls, *components: type[BaseElement], theme: TTheme | None = None
    ) -> Self:
        return cls(from_components(*components, theme=theme), type="text/css")

    @classmethod
    def load(cls, cache: bool = False, theme: TTheme | None = None) -> Self:
        return cls(from_loaded(cache=cache, theme=theme), type="text/css")

    def __getitem__(self, key: str | tuple[str, ...]) -> CSSProperties | GlobalStyles:
        return self.styles[key]

    def __iter__(self) -> Iterator[str | tuple[str, ...]]:
        return iter(self.styles.keys())

    def __len__(self) -> int:
        return len(self.styles)

    @property
    def styles(self) -> GlobalStyles:
        if isinstance(self.children[0], str):
            return {}
        elif callable(self.children[0]):
            return self.children[0](self.context["theme"])
        else:
            return self.children[0]

    @styles.setter
    def styles(self, value: GlobalStyles) -> None:
        self.children = (value,)

    def to_html(self) -> str:
        attributes = ""
        if formatted_attrs := self._format_attributes():
            attributes = f" {formatted_attrs}"

        if isinstance(self.children[0], str):
            css_styles = self.children[0]
        else:
            css_styles = format_styles(self.styles)

        return (
            f"<{self.html_name}{attributes}>\n"
            f"{css_styles}\n"
            f"</{self.html_name}>"
        )  # fmt: off


class script(Element[PrimitiveChildren, ScriptAttrs]):
    html_name = "script"

    def __init__(
        self, *children: PrimitiveChildren, **attrs: Unpack[ScriptAttrs]
    ) -> None:
        super().__init__(*children, **attrs)


class noscript(Element[AnyChildren, GlobalAttrs]):
    html_name = "noscript"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class meta(Element[NoChildren, MetaAttrs]):
    void_element = True
    html_name = "meta"

    def __init__(self, *children: NoChildren, **attrs: Unpack[MetaAttrs]) -> None:
        super().__init__(*children, **attrs)


class head(Element[AnyChildren, NoAttrs]):
    html_name = "head"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[NoAttrs]) -> None:
        super().__init__(*children, **attrs)


class body(Element[AnyChildren, GlobalAttrs]):
    html_name = "body"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class footer(Element[AnyChildren, GlobalAttrs]):
    html_name = "footer"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class html(ElementStrict[head, body, HtmlTagAttrs]):
    html_header = "<!doctype html>"
    html_name = "html"

    def __init__(
        self, *children: *tuple[head, body], **attrs: Unpack[HtmlTagAttrs]
    ) -> None:
        super().__init__(*children, **attrs)


class iframe(Element[NoChildren, IframeAttrs]):
    html_name = "iframe"

    def __init__(self, *children: NoChildren, **attrs: Unpack[IframeAttrs]) -> None:
        super().__init__(*children, **attrs)


class article(Element[AnyChildren, GlobalAttrs]):
    html_name = "article"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class address(Element[AnyChildren, GlobalAttrs]):
    html_name = "address"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class caption(Element[AnyChildren, GlobalAttrs]):
    html_name = "caption"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class col(Element[NoChildren, ColAttrs]):
    void_element = True
    html_name = "col"

    def __init__(self, *children: NoChildren, **attrs: Unpack[ColAttrs]) -> None:
        super().__init__(*children, **attrs)


class colgroup(Element[AnyChildren, GlobalAttrs]):
    html_name = "colgroup"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class area(Element[NoChildren, AreaAttrs]):
    void_element = True
    html_name = "area"

    def __init__(self, *children: NoChildren, **attrs: Unpack[AreaAttrs]) -> None:
        super().__init__(*children, **attrs)


class aside(Element[AnyChildren, GlobalAttrs]):
    html_name = "aside"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class source(Element[NoChildren, SourceAttrs]):
    void_element = True
    html_name = "source"

    def __init__(self, *children: NoChildren, **attrs: Unpack[SourceAttrs]) -> None:
        super().__init__(*children, **attrs)


class audio(Element[AnyChildren, AudioAttrs]):
    html_name = "audio"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[AudioAttrs]) -> None:
        super().__init__(*children, **attrs)


class base(Element[NoChildren, BaseAttrs]):
    void_element = True
    html_name = "base"

    def __init__(self, *children: NoChildren, **attrs: Unpack[BaseAttrs]) -> None:
        super().__init__(*children, **attrs)


class bdi(Element[AnyChildren, GlobalAttrs]):
    html_name = "bdi"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class bdo(Element[AnyChildren, GlobalAttrs]):
    html_name = "bdo"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class canvas(Element[AnyChildren, CanvasAttrs]):
    html_name = "canvas"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[CanvasAttrs]) -> None:
        super().__init__(*children, **attrs)


class data(Element[AnyChildren, DataAttrs]):
    html_name = "data"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[DataAttrs]) -> None:
        super().__init__(*children, **attrs)


class datalist(Element[AnyChildren, GlobalAttrs]):
    html_name = "datalist"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class details(Element[AnyChildren, DetailsAttrs]):
    html_name = "details"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[DetailsAttrs]) -> None:
        super().__init__(*children, **attrs)


class dfn(Element[AnyChildren, GlobalAttrs]):
    html_name = "dfn"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class dialog(Element[AnyChildren, DialogAttrs]):
    html_name = "dialog"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[DialogAttrs]) -> None:
        super().__init__(*children, **attrs)


class embed(Element[NoChildren, EmbedAttrs]):
    void_element = True
    html_name = "embed"

    def __init__(self, *children: NoChildren, **attrs: Unpack[EmbedAttrs]) -> None:
        super().__init__(*children, **attrs)


class figcaption(Element[AnyChildren, GlobalAttrs]):
    html_name = "figcaption"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class figure(Element[AnyChildren, GlobalAttrs]):
    html_name = "figure"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class hrgroup(Element[AnyChildren, GlobalAttrs]):
    html_name = "hrgroup"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class hr(Element[NoChildren, GlobalAttrs]):
    void_element = True
    html_name = "hr"

    def __init__(self, *children: NoChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class kbd(Element[AnyChildren, GlobalAttrs]):
    html_name = "kbd"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class map(Element[AnyChildren, MapAttrs]):
    html_name = "map"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[MapAttrs]) -> None:
        super().__init__(*children, **attrs)


class menu(Element[AnyChildren, GlobalAttrs]):
    html_name = "menu"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class meter(Element[AnyChildren, MeterAttrs]):
    html_name = "meter"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[MeterAttrs]) -> None:
        super().__init__(*children, **attrs)


class nav(Element[AnyChildren, GlobalAttrs]):
    html_name = "nav"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class object(Element[AnyChildren, ObjectAttrs]):
    html_name = "object"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[ObjectAttrs]) -> None:
        super().__init__(*children, **attrs)


class param(Element[NoChildren, ParamAttrs]):
    void_element = True
    html_name = "param"

    def __init__(self, *children: NoChildren, **attrs: Unpack[ParamAttrs]) -> None:
        super().__init__(*children, **attrs)


class picture(Element[AnyChildren, GlobalAttrs]):
    html_name = "picture"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class progress(Element[AnyChildren, ProgressAttrs]):
    html_name = "progress"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[ProgressAttrs]) -> None:
        super().__init__(*children, **attrs)


class q(Element[AnyChildren, QAttrs]):
    html_name = "q"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[QAttrs]) -> None:
        super().__init__(*children, **attrs)


class rp(Element[AnyChildren, GlobalAttrs]):
    html_name = "rp"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class rt(Element[AnyChildren, GlobalAttrs]):
    html_name = "rt"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class ruby(Element[AnyChildren, GlobalAttrs]):
    html_name = "ruby"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class samp(Element[AnyChildren, GlobalAttrs]):
    html_name = "samp"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class search(Element[AnyChildren, GlobalAttrs]):
    html_name = "search"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class sub(Element[AnyChildren, GlobalAttrs]):
    html_name = "sub"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class summary(Element[AnyChildren, GlobalAttrs]):
    html_name = "summary"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class sup(Element[AnyChildren, GlobalAttrs]):
    html_name = "sup"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class template(Element[AnyChildren, HtmlAttrs]):
    html_name = "template"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[HtmlAttrs]) -> None:
        super().__init__(*children, **attrs)


class time(Element[AnyChildren, TimeAttrs]):
    html_name = "time"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[TimeAttrs]) -> None:
        super().__init__(*children, **attrs)


class track(Element[NoChildren, TrackAttrs]):
    void_element = True
    html_name = "track"

    def __init__(self, *children: NoChildren, **attrs: Unpack[TrackAttrs]) -> None:
        super().__init__(*children, **attrs)


class var(Element[AnyChildren, GlobalAttrs]):
    html_name = "var"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)


class video(Element[AnyChildren, VideoAttrs]):
    html_name = "video"

    def __init__(self, *children: AnyChildren, **attrs: Unpack[VideoAttrs]) -> None:
        super().__init__(*children, **attrs)


class wbr(Element[NoChildren, GlobalAttrs]):
    void_element = True
    html_name = "wbr"

    def __init__(self, *children: NoChildren, **attrs: Unpack[GlobalAttrs]) -> None:
        super().__init__(*children, **attrs)



from ludic.catalog.typography import Link, Paragraph
from ludic.format import format_attr_value, format_attrs
from ludic.html import b, div, i, p, strong


def test_format_attr_value() -> None:
    assert format_attr_value("foo", "bar") == "bar"
    assert format_attr_value("foo", "bar", is_html=True) == "bar"

    assert format_attr_value("foo", 1) == "1"
    assert format_attr_value("foo", 1, is_html=True) == "1"

    assert format_attr_value("foo", True) == "true"
    assert format_attr_value("foo", True, is_html=True) == "foo"

    assert format_attr_value("foo", False) == "false"
    assert format_attr_value("foo", False, is_html=True) == ""

    assert format_attr_value("hx-confirm", True) == "true"
    assert format_attr_value("hx-confirm", True, is_html=True) == "true"

    assert format_attr_value("hx-confirm", False) == "false"
    assert format_attr_value("hx-confirm", False, is_html=True) == "false"

    assert (
        format_attr_value("style", {"color": "red", "background": "blue"})
        == "color:red;background:blue"
    )


def test_template_processing() -> None:
    # With Python 3.14+, t-strings are processed into tuples of parts
    # Example: t"test {b("foo")} {i("bar")}" becomes ("test ", b("foo"), " ", i("bar"))
    assert div(t"test {b('foo')} {i('bar')}") == div("test ", b("foo"), " ", i("bar"))


def test_tstring_in_elements() -> None:
    # Simple t-string with elements
    text = t"test {b('baz')} {i('foo')}"
    assert div(text) == div("test ", b("baz"), " ", i("foo"))

    # Nested t-strings
    assert div(t"test {div(t'foo {b("nested")}, {i("nested2")}')}") == div(
        "test ",
        div("foo ", b("nested"), ", ", i("nested2")),
    )


def test_component_with_tstring() -> None:
    paragraph = Paragraph(
        t"Hello, how {strong('are you')}? Click {Link('here', to='https://example.com')}.",
    )
    assert len(paragraph.children) == 5
    assert isinstance(paragraph.children[3], Link)
    assert paragraph.children[3].attrs["to"] == "https://example.com"
    assert paragraph.to_string(pretty=False) == (
        "<Paragraph>"
          'Hello, how <strong>are you</strong>? '
          'Click <Link to="https://example.com">here</Link>.'
        "</Paragraph>"
    )  # fmt: skip
    assert paragraph.to_html() == (
        "<p>"
        "Hello, how <strong>are you</strong>? Click "
        '<a href="https://example.com" target="_blank">here</a>.'
        "</p>"
    )


def test_escaping_works() -> None:
    """Test that HTML escaping works correctly with t-strings."""
    link = '<a href="https://example.com">test</a>'
    dom = p(t"Hello, how <b>are you</b>? Click {link}.")
    assert dom.to_html() == (
        "<p>Hello, how &lt;b&gt;are you&lt;/b&gt;? "
        'Click &lt;a href="https://example.com"&gt;test&lt;/a&gt;.</p>'
    )


def test_quotes_not_escaped() -> None:
    dom = p("It's alive <3.")
    assert dom.to_html() == "<p>It's alive &lt;3.</p>"


def test_attributes() -> None:
    assert format_attrs({"checked": False}, is_html=True) == {}
    assert format_attrs({"checked": True}, is_html=True) == {"checked": "checked"}

    assert format_attrs({"on_click": "test"}) == {"onclick": "test"}
    assert format_attrs({"hx_boost": True}) == {"hx-boost": "true"}
    assert format_attrs({"for_": "value"}) == {"for": "value"}
    assert format_attrs({"class_": "a b c"}) == {"class": "a b c"}
    assert format_attrs({"class_": "a b c", "classes": ["more", "classes"]}) == {
        "class": "a b c more classes"
    }

    assert format_attrs(
        {
            "hx_on_htmx_before_request": "alert('test')",
            "hx_on__after_request": "alert('test2')",
        }
    ) == {
        "hx-on-htmx-before-request": "alert('test')",
        "hx-on--after-request": "alert('test2')",
    }
    assert format_attrs(
        {"hx-on:htmx:before-request": "alert('Making a request!')"}
    ) == {"hx-on:htmx:before-request": "alert('Making a request!')"}



from fastapi import FastAPI, Response
from fastapi.testclient import TestClient

from ludic.attrs import GlobalAttrs
from ludic.catalog.layouts import Center, Stack
from ludic.components import Blank, Component
from ludic.contrib.fastapi import LudicRoute
from ludic.html import div, p, span
from ludic.types import AnyChildren
from ludic.web import Request

app = FastAPI()
app.router.route_class = LudicRoute


class MyTestComponent(Component[AnyChildren, GlobalAttrs]):
    classes = ["test"]
    styles = {".test": {"display": "block"}}

    def render(self) -> Center:
        return Center(
            Stack(*self.children, **self.attrs),
            style={"padding-block": self.theme.sizes.l},
        )


@app.get("/component")
async def return_component() -> div:
    return div(span("Hello"), p("World"))


@app.get("/dict", response_model=None)
async def return_dict() -> dict[str, str]:
    return {"message": "Hello World"}


@app.get("/custom-response")
async def return_custom_response() -> Response:
    return Response(content="Custom", media_type="text/plain")


@app.get("/nested")
async def nested_components() -> div:
    return div(p("Level 1", span("Level 2", div("Level 3"))), id="root")


@app.get("/t-string")
async def t_string_test() -> span:
    return span(t"Result: {p('This is a test')}")


@app.get("/status-code", status_code=202)
async def status_code_test() -> span:
    return span("this is a test")


@app.get("/blank-component")
async def blank_component() -> Blank[span]:
    return Blank(span("this is a test"))


@app.get("/url-for-call")
async def url_for_call(request: Request) -> span:
    return span(str(request.url_for(blank_component)))


@app.get("/use-component")
async def use_component() -> MyTestComponent:
    return MyTestComponent(span("inside-span"), id="test-component-1234")


def test_auto_component_conversion() -> None:
    client = TestClient(app)
    response = client.get("/component")

    assert response.status_code == 200
    assert response.content == b"<div><span>Hello</span><p>World</p></div>"
    assert response.headers["content-type"] == "text/html; charset=utf-8"


def test_non_component_responses() -> None:
    client = TestClient(app)

    response = client.get("/dict")
    assert response.status_code == 200
    assert response.json() == {"message": "Hello World"}

    response = client.get("/custom-response")
    assert response.status_code == 200
    assert response.content == b"Custom"
    assert response.headers["content-type"] == "text/plain; charset=utf-8"


def test_nested_components() -> None:
    client = TestClient(app)
    response = client.get("/nested")

    assert response.status_code == 200
    assert (
        b'<div id="root"><p>Level 1<span>Level 2<div>Level 3</div></span></p></div>'
        in response.content
    )


def test_formatter_cleanup_with_t_string() -> None:
    client = TestClient(app)
    response = client.get("/t-string")

    assert response.status_code == 200
    assert b"Result: <p>This is a test</p>" in response.content


def test_return_different_status_code() -> None:
    client = TestClient(app)
    response = client.get("/status-code")

    assert response.status_code == 202


def test_blank_component() -> None:
    client = TestClient(app)
    response = client.get("/blank-component")

    assert response.status_code == 200


def test_url_for_call() -> None:
    client = TestClient(app)
    response = client.get("/url-for-call")

    assert response.status_code == 200


def test_use_component() -> None:
    client = TestClient(app)
    response = client.get("/use-component")

    assert response.status_code == 200
    assert b"test-component-1234" in response.content
